@@ -45,14 +45,14 @@ from cloud_builder.version import __version__
 from cloud_builder.cloud_logger import CBCloudLogger
 from cloud_builder.response.response import CBResponse
 from cloud_builder.defaults import Defaults
-from cloud_builder.package_metadata.package_metadata import CBPackageMetaData
+from cloud_builder.project_metadata.project_metadata import CBProjectMetaData
 from cloud_builder.broker import CBMessageBroker
 from kiwi.command import Command
 from kiwi.privileges import Privileges
 from kiwi.path import Path
 from apscheduler.schedulers.background import BlockingScheduler
 from typing import (
-    Dict, Any
+    Dict, List, NamedTuple, Optional
 )
 
 from cloud_builder.exceptions import (
@@ -60,14 +60,22 @@ from cloud_builder.exceptions import (
     CBSchedulerIntervalError
 )
 
+request_validation_type = NamedTuple(
+    'request_validation_type', [
+        ('project_config', Dict),
+        ('response', Optional[CBResponse]),
+        ('is_valid', bool)
+    ]
+)
+
 
 @exception_handler
 def main() -> None:
     """
-    cb-scheduler - listens on incoming package build requests
-    from the message broker on a regular schedule. Only if
-    the max package to build limit is not exceeded, request
-    messages from the broker are accepted. In case the request
+    cb-scheduler - listens on incoming package and image build
+    requests from the message broker on a regular schedule.
+    Only if the max package/image to build limit is not exceeded,
+    request messages from the broker are accepted. In case the request
     matches the runner capabilities e.g architecture it gets
     acknowledged and removed from the broker queue.
 
@@ -120,6 +128,46 @@ def main() -> None:
     The .kiwi metadata file has to provide instructions
     such that the creation of a correct buildroot for these
     profiles is possible.
+
+    --
+
+    An image can be build for different profiles, build arguments
+    and architectures. An example image config to build myimage
+    for myprofile and for the x86_64 achitecture would look like
+    the following:
+
+    .. code:: yaml
+
+        schema_version: 0.1
+        name: myimage
+
+        images:
+          -
+            arch: x86_64
+            runner_group: suse
+            selection:
+              name: standard
+              profiles:
+                - myprofile
+              build_arguments:
+                - "--clear-cache"
+
+    The above instructs the scheduler to build one image for the
+    myprofile profile and the x86_64 achitecture on a runner in the
+    suse group. The package cache on this runner will be cleared
+    prior building the image. The image output files will get pacakged
+    into an rpm package. To do this properly the scheduler creates a
+    script which calls cb-image.
+
+    The directory containing the image config file:
+
+        Defaults.get_cloud_builder_metadata_file_name()
+
+    is treated as the image description and passed as such to the
+    KIWI image builder via cb-image. KIWI searches for a *.kiwi file
+    to accept the directory as an image description. If the cloud
+    builder image config file names a profile, that profile must be
+    defined in the KIWI *.kiwi file
     """
     args = docopt(
         __doc__,
@@ -133,7 +181,7 @@ def main() -> None:
     log.set_logfile()
 
     Path.create(
-        Defaults.get_runner_package_root()
+        Defaults.get_runner_root()
     )
 
     running_limit = int(args['--package-limit'] or 10)
@@ -163,8 +211,8 @@ def handle_build_requests(
 ) -> None:
     """
     Check on the runner state and if ok listen to the
-    message broker queue for new package build requests
-    The package_request_queue is used as shared queue
+    message broker queue for new package/image build requests
+    The build_request_queue is used as shared queue
     within a single group. It's important to have this
     queue configured to distribute messages across
     several readers to let multiple CB scheduler scale
@@ -195,29 +243,70 @@ def handle_build_requests(
             for message in broker.read(
                 topic=broker.get_runner_group(), timeout_ms=poll_timeout
             ):
-                request = broker.validate_package_request(message.value)
+                request = broker.validate_build_request(message.value)
                 if request:
-                    package_source_path = os.path.join(
+                    project_source_path = os.path.join(
                         Defaults.get_runner_project_dir(),
                         format(request['project'])
                     )
-                    package_config = is_request_valid(
-                        package_source_path, request, log, broker
+                    validated_request = is_request_valid(
+                        project_source_path, request, log
                     )
-                    if package_config:
+                    if validated_request.is_valid:
+                        broker.acknowledge()
+
+                    if validated_request.response:
+                        log.response(validated_request.response, broker)
+
+                    build_config = validated_request.project_config
+
+                    if build_config and 'distributions' in build_config:
                         build_package(request, broker, log)
+                    elif build_config and 'images' in build_config:
+                        build_image(request, build_config, broker, log)
     finally:
         log.info('Closing message broker connection')
         broker.close()
 
 
+def build_image(
+    request: Dict, build_config: Dict, broker: CBMessageBroker,
+    log: CBCloudLogger
+) -> None:
+    """
+    Update the image sources and run the script which
+    utilizes cb-image to build the image for the requested
+    target
+
+    :param dict request: yaml dict request record
+    :param dict build_config: yaml dict project configuration
+    :param CBMessageBroker broker: instance of CBMessageBroker
+    """
+    log.set_id(os.path.basename(request['project']))
+    reset_build_if_running(
+        request, log, broker
+    )
+    status_flags = Defaults.get_status_flags()
+    if request['action'] == status_flags.image_rebuild or \
+       request['action'] == status_flags.image_source_rebuild:
+        log.info('Update project git source repo prior build')
+        Command.run(
+            ['git', '-C', Defaults.get_runner_project_dir(), 'pull']
+        )
+
+    log.info('Starting image build process')
+    Command.run(
+        ['bash', create_image_run_script(request, build_config)]
+    )
+
+
 def build_package(
     request: Dict, broker: CBMessageBroker, log: CBCloudLogger
 ) -> None:
     """
     Update the package sources and run the script which
     utilizes cb-prepare/cb-run to build the package for
-    all configured targets
+    the requested target
 
     :param dict request: yaml dict request record
     :param CBMessageBroker broker: instance of CBMessageBroker
@@ -241,9 +330,9 @@ def build_package(
        request['action'] == status_flags.package_rebuild_clean:
         buildroot_rebuild = True
 
-    log.info('Starting build process')
+    log.info('Starting package build process')
     Command.run(
-        ['bash', create_run_script(request, buildroot_rebuild)]
+        ['bash', create_package_run_script(request, buildroot_rebuild)]
     )
 
 
@@ -258,23 +347,18 @@ def reset_build_if_running(
     :param CBCloudLogger log: logger instance
     :param CBMessageBroker broker: instance of CBMessageBroker
     """
-    package_root = os.path.join(
-        Defaults.get_runner_package_root(), request['project']
+    build_pid = 0
+    status_flags = Defaults.get_status_flags()
+    response = CBResponse(request['request_id'], log.get_id())
+    project_root = os.path.join(
+        Defaults.get_runner_root(), request['project']
     )
-    dist_profile = f'{request["package"]["dist"]}.{request["package"]["arch"]}'
-    build_root = f'{package_root}@{dist_profile}'
-    package_run_pid = f'{build_root}.pid'
-    if os.path.isfile(package_run_pid):
-        with open(package_run_pid) as pid_fd:
-            build_pid = int(pid_fd.read().strip())
-        log.info(
-            'Checking state of former build with PID:{0}'.format(
-                build_pid
-            )
-        )
-        if psutil.pid_exists(build_pid):
-            status_flags = Defaults.get_status_flags()
-            response = CBResponse(request['request_id'], log.get_id())
+    if 'package' in request:
+        dist_profile = \
+            f'{request["package"]["dist"]}.{request["package"]["arch"]}'
+        build_root = f'{project_root}@{dist_profile}'
+        build_pid = is_active(f'{build_root}.pid', log)
+        if build_pid:
             response.set_package_jobs_reset_response(
                 message='Kill job group for PID:{0} prior rebuild'.format(
                     build_pid
@@ -284,8 +368,24 @@ def reset_build_if_running(
                 arch=request['package']['arch'],
                 dist=request['package']['dist']
             )
-            log.response(response, broker)
-            os.kill(build_pid, signal.SIGTERM)
+    elif 'image' in request:
+        selection = \
+            f'{request["image"]["selection"]}.{request["image"]["arch"]}'
+        image_root = f'{project_root}@{selection}'
+        build_pid = is_active(f'{image_root}.pid', log)
+        if build_pid:
+            response.set_image_jobs_reset_response(
+                message='Kill job group for PID:{0} prior rebuild'.format(
+                    build_pid
+                ),
+                response_code=status_flags.reset_running_build,
+                image=request['project'],
+                arch=request['image']['arch'],
+                selection=request['image']['selection']
+            )
+    if build_pid:
+        log.response(response, broker)
+        os.kill(build_pid, signal.SIGTERM)
 
 
 def get_running_builds() -> int:
@@ -301,103 +401,268 @@ def get_running_builds() -> int:
 
 
 def is_request_valid(
-    package_source_path: str, request: Dict, log: CBCloudLogger, broker: Any
-) -> Dict:
+    project_source_path: str, request: Dict, log: CBCloudLogger = None
+) -> request_validation_type:
     """
     Sanity checks between the request and the package sources
 
-    1. Check if given package source exists
-    2. Check if there is a cloud builder metadata and kiwi file
-    3. Check if architecture + dist + runner_group is configured in the metadata
-    4. Check if the host architecture is compatbile with the
+    1. Check if given package/image source exists
+    2. Check if there is a cloud builder metadata and a .kiwi file
+    3. Check if architecture + runner_group is configured in the metadata
+    4. Check if dist is configured for package builds
+    5. Check if the host architecture is compatbile with the
        request architecture
 
-    :param str package_source_path: path to package sources
+    :param str project_source_path: path to package/image sources
     :param dict request: yaml dict request record
-    :param CBCloudLogger log: logger instance
-    :param CBMessageBroker broker: instance of CBMessageBroker
 
-    :return: metadata config dict
+    :return: request_validation_type
 
-    :rtype: dict
+    :rtype: tuple
     """
     status_flags = Defaults.get_status_flags()
     response = CBResponse(
-        request['request_id'], log.get_id()
+        request['request_id'], log.get_id() if log else 'local'
     )
-    # 1. Check on package sources to exist
-    if not os.path.isdir(package_source_path):
+    # 1. Check on project sources to exist
+    if not os.path.isdir(project_source_path):
         response.set_project_not_existing_response(
-            message=f'Package does not exist: {package_source_path}',
-            response_code=status_flags.package_not_existing,
+            message=f'Project does not exist: {project_source_path}',
+            response_code=status_flags.project_not_existing,
             project=request['project']
         )
-        log.response(response, broker)
-        return {}
+        return request_validation_type(
+            project_config={},
+            response=response,
+            is_valid=False
+        )
 
-    # 2. Check on package metadata to exist
-    package_metadata = os.path.join(
-        package_source_path, Defaults.get_cloud_builder_metadata_file_name()
+    # 2. Check on project metadata to exist
+    project_metadata = os.path.join(
+        project_source_path, Defaults.get_cloud_builder_metadata_file_name()
     )
-    if not os.path.isfile(package_metadata):
+    if not os.path.isfile(project_metadata):
         response.set_project_metadata_not_existing_response(
-            message=f'Package metadata does not exist: {package_metadata}',
-            response_code=status_flags.package_metadata_not_existing,
+            message=f'Project metadata does not exist: {project_metadata}',
+            response_code=status_flags.project_metadata_not_existing,
             project=request['project']
         )
-        log.response(response, broker)
-        return {}
+        return request_validation_type(
+            project_config={},
+            response=response,
+            is_valid=False
+        )
 
-    # 3. Check if requested arch+dist is configured
-    target_ok = False
-    request_arch = request['package']['arch']
-    request_dist = request['package']['dist']
-    request_runner_group = request['runner_group']
-    package_config = CBPackageMetaData.get_package_config(
-        package_source_path, log, request['request_id']
+    # 3. Check if requested package arch+dist+runner_group is configured
+    project_config = CBProjectMetaData.get_project_config(
+        project_source_path, log, request['request_id']
     )
-    if package_config:
-        for target in package_config.get('distributions') or []:
+    if not project_config:
+        return request_validation_type(
+            project_config={},
+            response=None,
+            is_valid=False
+        )
+    request_arch = 'none'
+    request_dist = 'none'
+    request_selection = 'none'
+    if 'package' in request:
+        target_ok = False
+        request_arch = request['package']['arch']
+        request_dist = request['package']['dist']
+        request_runner_group = request['runner_group']
+        for target in project_config.get('distributions') or []:
+            allowed_runner_groups = [target.get('runner_group'), 'local']
             if request_arch == target.get('arch') and \
                request_dist == target.get('dist') and \
-               request_runner_group == target.get('runner_group'):
+               request_runner_group in allowed_runner_groups:
                 target_ok = True
                 break
-    if not target_ok:
-        response.set_project_invalid_target_response(
-            message='No build config for: {0}.{1} in runner group {2}'.format(
-                request_dist, request_arch, request_runner_group
-            ),
-            response_code=status_flags.package_target_not_configured,
-            project=request['project']
-        )
-        log.response(response, broker)
-        return {}
+        if not target_ok:
+            response.set_project_invalid_target_response(
+                message='No {0} config for: {1}.{2} in runner group {3}'.format(
+                    'package', request_dist, request_arch, request_runner_group
+                ),
+                response_code=status_flags.package_target_not_configured,
+                project=request['project']
+            )
+            return request_validation_type(
+                project_config={},
+                response=response,
+                is_valid=False
+            )
 
-    # 4. Check if host architecture is compatbile
-    if not request['package']['arch'] == platform.machine():
+    # 4. Check if requested image arch+runner_group+selection is configured
+    if 'image' in request:
+        target_ok = False
+        request_arch = request['image']['arch']
+        request_selection = request['image']['selection']
+        request_runner_group = request['runner_group']
+        for target in project_config.get('images') or []:
+            allowed_runner_groups = [target.get('runner_group'), 'local']
+            if request_arch == target.get('arch') and \
+               request_selection == target.get('selection').get('name') and \
+               request_runner_group in allowed_runner_groups:
+                target_ok = True
+                break
+        if not target_ok:
+            response.set_project_invalid_target_response(
+                message='No {0} config for: {1}.{2} in runner group {3}'.format(
+                    'image', request_selection, request_arch,
+                    request_runner_group
+                ),
+                response_code=status_flags.image_target_not_configured,
+                project=request['project']
+            )
+            return request_validation_type(
+                project_config={},
+                response=response,
+                is_valid=False
+            )
+
+    # 5. Check if host architecture is compatbile
+    if request_arch and request_arch != platform.machine():
         response.set_buildhost_arch_incompatible_response(
             message=f'Incompatible arch: {platform.machine()}',
-            response_code=status_flags.package_request_accepted,
+            response_code=status_flags.incompatible_build_arch,
             package=request['project']
         )
-        log.response(response, broker)
-        return {}
+        return request_validation_type(
+            project_config={},
+            response=response,
+            is_valid=False
+        )
 
     # All good...
-    response.set_package_build_scheduled_response(
-        message='Accept package build request',
-        response_code=status_flags.package_request_accepted,
-        package=request['project'],
-        arch=request['package']['arch'],
-        dist=request['package']['dist']
+    if 'package' in request:
+        response.set_package_build_scheduled_response(
+            message='Accept package build request',
+            response_code=status_flags.package_request_accepted,
+            package=request['project'],
+            arch=request_arch,
+            dist=request_dist
+        )
+    elif 'image' in request:
+        response.set_image_build_scheduled_response(
+            message='Accept image build request',
+            response_code=status_flags.image_request_accepted,
+            image=request['project'],
+            arch=request_arch,
+            selection=request_selection
+        )
+    return request_validation_type(
+        project_config=project_config,
+        response=response,
+        is_valid=True
     )
-    log.response(response, broker)
-    broker.acknowledge()
-    return package_config
 
 
-def create_run_script(
+def create_image_run_script(
+    request: Dict, build_config: Dict,
+    bundle_id: str = '0', local_build: bool = False
+) -> str:
+    """
+    Create script to call cb-image
+
+    :param dict request: yaml dict request record
+    :param dict build_config: yaml dict project configuration
+    :param str bundle_id: optional package bundle ID, defaults to '0'
+    :param bool local_build:
+        create script for build on localhost. This keeps
+        the script in the foreground and prints all information
+        to stdout instead of writing log files
+
+    :return: script file path name
+
+    :rtype: str
+    """
+    build_options: List[str] = []
+    profiles: List[str] = []
+    for target in build_config.get('images') or []:
+        selection = target['selection']
+        if selection['name'] == request['image']['selection']:
+            build_options = selection.get('build_arguments') or []
+            profiles = selection.get('profiles') or []
+
+    profile_opts = []
+    for profile in profiles:
+        profile_opts.extend(['--profile', profile])
+
+    selection = request['image']['selection']
+
+    custom_args = ['--']
+    for argument in build_options:
+        custom_args.append(argument)
+
+    if local_build:
+        image_source_path = request['project']
+        image_target_path = \
+            f'{image_source_path}@{selection}.{request["image"]["arch"]}'
+
+        run_script = dedent('''
+            #!/bin/bash
+            set -e
+            rm -rf {image_target_path}
+            cb-image \\
+                --request-id {request_id} \\
+                --bundle-id {bundle_id} \\
+                --description {image_source_path} \\
+                --target-dir {image_target_path} \\
+                --local \\
+                {profile_opts} {custom_args}
+        ''').format(
+            image_source_path=image_source_path,
+            image_target_path=image_target_path,
+            profile_opts=' '.join(profile_opts) if profile_opts else '',
+            custom_args=' '.join(custom_args) if build_options else '',
+            request_id=request['request_id'],
+            bundle_id=bundle_id
+        )
+    else:
+        image_source_path = os.path.join(
+            Defaults.get_runner_project_dir(), request['project']
+        )
+        image_target_path = '{0}@{1}.{2}'.format(
+            os.path.join(Defaults.get_runner_root(), request['project']),
+            selection, request['image']['arch']
+        )
+        run_script = dedent('''
+            #!/bin/bash
+            set -e
+
+            rm -rf {image_target_path}
+
+            function finish {{
+                kill $(jobs -p) &>/dev/null
+            }}
+
+            {{
+                trap finish EXIT
+                cb-image \\
+                    --request-id {request_id} \\
+                    --bundle-id {bundle_id} \\
+                    --description {image_source_path} \\
+                    --target-dir {image_target_path} \\
+                    {profile_opts} {custom_args}
+            }} &>>{image_target_path}.run.log &
+
+            echo $! > {image_target_path}.pid
+        ''').format(
+            image_source_path=image_source_path,
+            image_target_path=image_target_path,
+            profile_opts=' '.join(profile_opts) if profile_opts else '',
+            custom_args=' '.join(custom_args) if build_options else '',
+            request_id=request['request_id'],
+            bundle_id=bundle_id
+        )
+    image_run_script = f'{image_target_path}.sh'
+    Path.create(os.path.dirname(image_run_script))
+    with open(image_run_script, 'w') as script:
+        script.write(run_script)
+    return image_run_script
+
+
+def create_package_run_script(
     request: Dict, buildroot_rebuild: bool, local_build: bool = False
 ) -> str:
     """
@@ -449,7 +714,7 @@ def create_run_script(
             Defaults.get_runner_project_dir(), request['project']
         )
         package_root = os.path.join(
-            Defaults.get_runner_package_root(), request['project']
+            Defaults.get_runner_root(), request['project']
         )
         build_root = f'{package_root}@{dist_profile}'
         run_script = dedent('''
@@ -490,3 +755,17 @@ def create_run_script(
     with open(package_run_script, 'w') as script:
         script.write(run_script)
     return package_run_script
+
+
+def is_active(pid_file: str, log: CBCloudLogger) -> int:
+    if os.path.isfile(pid_file):
+        with open(pid_file) as pid_fd:
+            build_pid = int(pid_fd.read().strip())
+        log.info(
+            'Checking state of former build with PID:{0}'.format(
+                build_pid
+            )
+        )
+        if psutil.pid_exists(build_pid):
+            return build_pid
+    return 0
